using System;
using System.Collections;
using System.Globalization;
using System.Reflection;
using System.Security;
using System.Security.Permissions;
using System.Text;

namespace DataBroker.Dao
{
    /// <summary>
    /// Base class for all entities. One entity is related to a single 
    /// physical database table. Joins are supported by defining properties of the
    /// type <see cref="EntityCollection">EntityCollection</see>.  
    /// </summary>
    /// <remarks>It is required that each property match the column name in the
    /// table. It is highly recommended that each property be paired with a private
    /// member with an leading underscore, so to allow the compare functions to operate
    /// directly in the private member instead of the property. This allows for the
    /// properties to have side effects such as returning a valid value if it is Null.</remarks>
    [Serializable]
    [CLSCompliantAttribute(true)]
    public class Entity : ICloneable, IDisposable
    {
        /// <summary>
        /// Status for entity.
        /// </summary>
        public enum EntityStatusEnum
        {
            /// <summary>
            /// Instances of persistence-capable classes behave like any other 
            /// objects when they are created by the new operator. Field access 
            /// is not mediated for these objects, resulting in almost the same 
            /// access performance as for unenhanced classes. Transient instances are 
            /// not influenced by transactional methods like begin, commit, or 
            /// rollback except when they are reachable by other persistent objects.
            /// </summary>
            Transient,
            /// <summary>
            /// A call to PersistenceManager.makePersistent() assigns a new object ID 
            /// to the instance, and the state changes from transient to PersistentNew. 
            /// A PersistentNew object behaves like a transient object, but is flushed 
            /// to the datastore at commit or whenever the object state needs to be 
            /// synchronized with internal caches. Other objects can have a PersistentNew
            /// state if they 
            /// are reachable through persistent fields. They are made persistent, and 
            /// so on, until all reachable objects are either of a transient type or 
            /// PersistentNew. Between makePersistent() and a final commit, references 
            /// may change and reachable instances may become unreachable. The state of 
            /// these provisionally-persistent instances is also PersistentNew. 
            /// These objects behave like PersistentNew objects, but their state 
            /// may change during commit, if the provisionally-persistent object 
            /// isn't reachable anymore.
            /// </summary>
            PersistentNew,
            /// <summary>
            /// After successful completion of the transaction, the PersistentNew 
            /// object's state changes to Hollow. A Hollow object is a reference 
            /// that represents some data in the datastore, which is not yet in 
            /// memory or outdated. Usually, other clients or applications are 
            /// free to access or change data in the datastore referenced by 
            /// hollow objectsfor instance, in a different transaction or process
            /// Additionally, the contents 
            /// of such hollow objects are cleared after commit, to free memory. 
            /// If references were kept between hollow objects, the garbage 
            /// collector may not free a graph of objects. After some transactions, 
            /// a bunch of hollow objects might be accumulated.
            ///The hollow state cannot be interrogated by the application. 
            ///There is no Helper method, and access to fields may result 
            ///in a state change so that it is impossible to tell the previous state.
            /// </summary>
            Hollow,
            /// <summary>
            /// Whenever PersistentNew objects are deleted by a call to 
            /// PersistenceManager.deletePersistent(), their state changes 
            /// to persistent-new-deleted. There is no way to undelete an object, 
            /// which means that this state is final until transaction termination. 
            /// Accessing a field of a deleted object results in a UserException, 
            /// except when reading primary-key fields. After transaction termination 
            /// (commit or rollback), the deleted instance's state changes back to transient.
            /// </summary>
            PersistentNewDeleted,
            /// <summary>
            /// Objects transition from Hollow to PersistentClean after they 
            /// have been filled with valid data from the datastore, but only 
            /// when the data has not yet been modified in the current transaction. 
            /// The state is called clean because the object's data is consistent 
            /// with the corresponding data in the datastore. 
            /// Either those references may point to objects already in memory, 
            /// or the PersistenceManager creates new hollow objects. 
            /// PersistentClean objects might be collected by garbage collector 
            /// if the application no longer refers to the objects. The 
            /// PersistenceManager is required to cache the references to 
            /// objects and instantiate new Hollow objects in case the objects 
            /// were already garbage-collected.
            /// </summary>
            PersistentClean,
            /// <summary>
            /// Objects transition from either PersistentClean or Hollow 
            /// to PersistentDirty when one of their persistent fields is changed. 
            /// This state is called dirty because the in-memory data of the object 
            /// does not reflect the persistent data corresponding to the object in 
            /// the datastore. If the value assigned to a managed field is equal to 
            /// the previous value, the instance's state nevertheless changes to 
            /// PersistentDirty. Furthermore, an instance does not change 
            /// back to PersistentClean or Hollow when previous field values or 
            /// modifications are restored. Whenever transient objects are assigned 
            /// to PersistentClean or PersistentDirty objects, they do not 
            /// transition to PersistentNew at that moment.
            /// </summary>
            PersistentDirty,
            /// <summary>
            /// In a call to PersistentManager.deletePersistent(), 
            /// the state of the persistent instance is changed to 
            /// persistent-deleted. It is impossible to leave this state 
            /// except by terminating the transaction, which means that 
            /// there is no method to undo a DeletePersistent call except 
            /// by aborting the transaction. A successful transaction 
            /// termination makes a deleted instance transient. 
            /// An object cannot be made persistent again with the same 
            /// identity as long as it is deleted.
            /// </summary>
            PersistentDeleted
        }
        private EntityStatusEnum entityStatus;
        private Entity originalEntity;
        private bool _IsFlat = true;
        private int _Index = -1;
        private string _UnderlyingTableName = "";

        [NotInTable]
        public string UnderlyingTableName
        {
            get { return _UnderlyingTableName; }
            set { _UnderlyingTableName = value; }
        }
        [NotInTable]
        public int Index
        {
            get { return _Index; }
            set { _Index = value; }
        }

        public void SetFlat(bool isFlat)
        {
            _IsFlat = isFlat;
        }

        public bool IsFlat()
        {
            return _IsFlat;
        }

        /// <summary>
        /// Public constructor.
        /// </summary>
        public Entity()
        {
            //When first created, the Entity is Transient.
            entityStatus = EntityStatusEnum.Transient;
        }

        public virtual void JoinAll()
        {

            foreach (PropertyInfo pi in this.GetType().GetProperties())
            {
                if (pi.PropertyType == typeof(EntityCollection))
                {
                    object[] attr = pi.GetCustomAttributes(typeof (TypeContainedAttribute), true);
                    if (attr.Length > 0)
                    {
                        TypeContainedAttribute tca = attr[0] as TypeContainedAttribute;
                        if (tca != null)
                        {
                            pi.SetValue(this, new EntityCollection(Activator.CreateInstance(tca.TypeContainedInCollection)), null);
                            PropertyInfo inPi = GetType().GetProperty(pi.Name + "_");
                            if (inPi != null)
                            {
                                inPi.SetValue(this, Activator.CreateInstance(tca.TypeContainedInCollection), null);
                            }
                        }
                    }
                }
            }
            SetFlat(false);
        }

        /// <summary>
        /// Checks if entity was marked for deletion.
        /// </summary>
        /// <returns>True if entity is marked for deletion; false otherwise.</returns>
        public bool IsDeleted()
        {
            return (entityStatus == EntityStatusEnum.PersistentDeleted ||
                    entityStatus == EntityStatusEnum.PersistentNewDeleted);
        }
        /// <summary>
        /// Checks if entity is dirty (was changed).
        /// </summary>
        /// <returns>True if entity is dirty, false otherwise.</returns>
        public bool IsDirty()
        {
            return (GetStatus() == EntityStatusEnum.PersistentDirty);
        }

        /// <summary>
        /// Checks if entity is new.
        /// </summary>
        /// <returns>True if entity is new, false otherwise.</returns>
        public bool IsNew()
        {
            return (GetStatus() == EntityStatusEnum.PersistentNew);
        }

        /// <summary>
        /// Checks if entity is persistent.
        /// </summary>
        /// <returns>True if entity is persistent, false otherwise.</returns>
        public bool IsPersistent()
        {
            return (GetStatus() == EntityStatusEnum.PersistentClean ||
                    GetStatus() == EntityStatusEnum.PersistentDeleted ||
                    GetStatus() == EntityStatusEnum.PersistentDirty ||
                    GetStatus() == EntityStatusEnum.PersistentNew ||
                    GetStatus() == EntityStatusEnum.PersistentNewDeleted);
        }
        /// <summary>
        /// Returns current status for instance.
        /// </summary>
        /// <returns></returns>
        public EntityStatusEnum GetStatus()
        {
            if (entityStatus == EntityStatusEnum.PersistentClean)
            {
                if (HasChanged())
                {
                    entityStatus = EntityStatusEnum.PersistentDirty;
                }
            }
            if (entityStatus == EntityStatusEnum.Transient)
            {
                if (HasChanged())
                {
                    entityStatus = EntityStatusEnum.PersistentNew;
                }
            }
            return entityStatus;
        }
        /// <summary>
        /// Checks if column is primary key.
        /// </summary>
        /// <param name="member">Member Info</param>
        /// <returns>True if it is primary key; false otherwise.</returns>
        internal static bool IsColumnPrimaryKey(MemberInfo member)
        {
            foreach (object attribute in member.GetCustomAttributes(true))
            {
                if (attribute is IsPKAttribute)
                {
                    return true;
                }
            }
            return false;
        }
        /// <summary>
        /// Verifies if an object has a value that is considered Null from the
        /// database standpoint. For DateTime, int and Double this includes MinValue.
        /// </summary>
        /// <param name="value">Value to verify.</param>
        /// <returns>True if null.</returns>
        internal static bool IsNullOrZero(object value)
        {
            if (value is DateTime)
                return (DateTime)value == DateTime.MinValue;
            else if (value is int)
                return ((int)value == 0);
            else if (value is Double)
                return ((double)value == double.MinValue);
            else if (value is Guid)
                return ((Guid)value == Guid.Empty);
            else
                return value == null;
        }

        /// <summary>
        /// Marks instance as being read from database.
        /// </summary>
        public void SetPersistentClean()
        {
            foreach (PropertyInfo prop in GetType().GetProperties())
            {
                if (IsNullOrZero(prop.GetValue(this, null)) && IsColumnPrimaryKey(prop))
                {
                    entityStatus = EntityStatusEnum.Transient;
                    return;
                }
            }
            originalEntity = (Entity)CloneEntity();
            entityStatus = EntityStatusEnum.PersistentClean;
        }


        /// <summary>
        /// Marks instance as being read from database.
        /// </summary>
        public void SetPersistentNew()
        {
            foreach (PropertyInfo prop in GetType().GetProperties())
            {
                if (IsColumnPrimaryKey(prop))
                {
                    if (prop.PropertyType == typeof(Int32))
                    {
                        FieldInfo fi =
                            GetType().GetField("_" + prop.Name, BindingFlags.NonPublic | BindingFlags.Instance);
                        fi.SetValue(this, 0);
                    }
                    entityStatus = EntityStatusEnum.PersistentNew;
                    return;

                }
            }
        }


        /// <summary>
        /// Marks the entity instance as new and persisted.
        /// </summary>
        internal void MakePersistent()
        {
            entityStatus = EntityStatusEnum.PersistentNew;
        }

        /// <summary>
        /// Marks the entity instance as deleted and persisted.
        /// </summary>
        internal void DeletePersistent()
        {
            entityStatus = EntityStatusEnum.PersistentNewDeleted;
        }

        /// <summary>
        /// Returns the status to Transiente after a rollback.
        /// </summary>
        internal void Rollback()
        {
            if (entityStatus == EntityStatusEnum.PersistentNew ||
                entityStatus == EntityStatusEnum.PersistentNewDeleted)
                entityStatus = EntityStatusEnum.Transient;
        }

        /// <summary>
        /// Resets the status after a commit.
        /// </summary>
        internal void Commit()
        {
            switch (entityStatus)
            {
                case EntityStatusEnum.PersistentNew:
                    entityStatus = EntityStatusEnum.Hollow;
                    break;
                case EntityStatusEnum.PersistentNewDeleted:
                    entityStatus = EntityStatusEnum.Transient;
                    break;
            }
        }
        /// <summary>
        /// Marks instance as being read from database.
        /// </summary>
        public void SetHollow()
        {
            entityStatus = EntityStatusEnum.Hollow;
        }

        /// <summary>
        /// Marks instance as deleted.
        /// </summary>1
        public virtual void SetDelete()
        {
            if (entityStatus == EntityStatusEnum.PersistentClean || entityStatus == EntityStatusEnum.PersistentDirty)
                entityStatus = EntityStatusEnum.PersistentDeleted;
            else
                entityStatus = EntityStatusEnum.PersistentNewDeleted;
        }

        /// <summary>
        /// Obtains name of entity.
        /// </summary>
        /// <returns>Entity name.</returns>
        public virtual string EntityName()
        {
            return "Base Entity";
        }

        /// <summary>
        /// Returns entity as read from the database.
        /// </summary>
        /// <returns>Original entity.</returns>
        public Entity GetOriginal()
        {
            return originalEntity;
        }

        /// <summary>
        /// Checks each property in the entity, and returns
        /// as false soon as any one of them is different. If they 
        /// are all the same, return true.
        /// </summary>
        /// <returns></returns>
        public bool HasChanged()
        {
            Type t = GetType();
            PropertyInfo[] props = t.GetProperties();
            foreach (PropertyInfo pi in props)
            {
                if (!IsNotInTable(pi) && !pi.Name.Equals("Index", StringComparison.InvariantCultureIgnoreCase))
                {
                    if (originalEntity != null)
                    {
                        if (pi.GetValue(originalEntity, null) != null)
                        {
                            if (!pi.GetValue(originalEntity, null).Equals(pi.GetValue(this, null)))
                                return true;
                        }
                        else
                        {
                            if (pi.GetValue(this, null) != null)
                                return true;
                        }
                    }
                    else
                    {
                        if (pi.GetValue(this, null) != null)
                            return true;
                    }
                }
            }
            return false;
        }

        /// <summary>
        /// Checks if property was changed.
        /// </summary>
        /// <param name="property">Property name.</param>
        /// <remarks>Uses the inner private member, if possible. This allows
        /// for the developer to return values other than Null for display or
        /// binding purposes.</remarks>
        /// <returns>True if the property was changed, false otherwise.</returns>
        public bool HasChanged(string property)
        {
            Type t = GetType();
            PropertyInfo pi = t.GetProperty(property);
            object originalValue = pi.GetValue(originalEntity, null);
            object currentValue = pi.GetValue(this, null);
            try
            {
                FieldInfo fi = GetType().GetField("_" + pi.Name, BindingFlags.NonPublic | BindingFlags.Instance);
                originalValue = fi.GetValue(originalEntity);
                currentValue = fi.GetValue(this);
            }
            finally
            {
            }
            return (originalValue != currentValue);
        }
        /// <summary>
        /// Gets list of property that were changed.
        /// </summary>
        /// <returns></returns>
        public ArrayList GetChangedProperties()
        {
            ArrayList changed = new ArrayList();
            PropertyInfo[] pi = GetType().GetProperties();

            try
            {
                new PermissionSet(PermissionState.Unrestricted).Assert();
                foreach (PropertyInfo p in pi)
                {
                    if (p.PropertyType != typeof(EntityCollection))
                        try
                        {
                            object orgValue, newValue;
                            Type propertyType;
                            object[] customAttributes = p.GetCustomAttributes(typeof(TypeAttribute), false);
                            if (customAttributes.Length > 0)
                                propertyType = ((TypeAttribute)customAttributes[0]).TypeName;
                            else
                                propertyType = p.PropertyType;
                            orgValue = p.GetValue(originalEntity, null);
                            try
                            {
                                Type type = GetType();
                                FieldInfo fi = type.GetField("_" + p.Name, BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.IgnoreCase);
                                object theValue = null;
                                if (fi != null)
                                    theValue = fi.GetValue(originalEntity);
                                orgValue = theValue;
                            }
                            catch
                            {
                            }
                            if (orgValue != null)
                            {
                                if (propertyType == typeof(bool))
                                {
                                    orgValue = bool.Parse(orgValue.ToString());
                                }
                            }
                            newValue = p.GetValue(this, null);
                            if (newValue == null && orgValue == null)
                                continue;
                            if (newValue != null)
                                newValue = Parse(newValue, propertyType);
                            if (orgValue != null)
                                orgValue = Parse(orgValue, propertyType);
                            if (newValue == null || !newValue.Equals(orgValue))
                            {
                                ArrayList chgd = new ArrayList();
                                chgd.Add(p.Name);
                                chgd.Add(orgValue);
                                chgd.Add(newValue);
                                changed.Add(chgd);
                            }
                        }
                        finally
                        {
                        }
                }
            }
            catch
            {
                //				Log.Error(ex.Message);
                throw;
            }
            finally
            {
                CodeAccessPermission.RevertAssert();
            }

            return changed;
        }


        /// <summary>
        /// Gets list of property that were changed.
        /// </summary>
        /// <returns></returns>
        public void Revert()
        {
            PropertyInfo[] pi = GetType().GetProperties();
            try
            {
                new PermissionSet(PermissionState.Unrestricted).Assert();
                foreach (PropertyInfo p in pi)
                {
                    if (p.PropertyType != typeof(EntityCollection))
                        try
                        {
                            try
                            {
                                Type type = GetType();
                                FieldInfo fi =
                                    type.GetField("_" + p.Name,
                                                  BindingFlags.NonPublic | BindingFlags.Instance |
                                                  BindingFlags.IgnoreCase);
                                if (fi != null)
                                {
                                    object theValue = fi.GetValue(originalEntity);
                                    fi.SetValue(this, theValue);
                                }
                            }
                            catch
                            {
                            }
                            finally
                            {
                            }
                        }
                        finally
                        {
                        }
                }
            }
            finally
            {
                entityStatus = EntityStatusEnum.PersistentClean;
                CodeAccessPermission.RevertAssert();
            }
        }


        /// <summary>
        /// Gets list of property that were changed.
        /// </summary>
        /// <returns></returns>
        public ArrayList GetChangedProperties(string[] columnsToChange)
        {
            ArrayList changed = new ArrayList();

            try
            {
                new PermissionSet(PermissionState.Unrestricted).Assert();
                foreach (string propName in columnsToChange)
                {
                    PropertyInfo pi = GetType().GetProperty(propName, BindingFlags.IgnoreCase | BindingFlags.Public | BindingFlags.Instance);
                    if (pi != null)
                    {
                        ArrayList chgd = new ArrayList();
                        chgd.Add(pi.Name);
                        chgd.Add(null);
                        chgd.Add(pi.GetValue(this, null));
                        changed.Add(chgd);
                    }
                }
            }
            finally
            {
                CodeAccessPermission.RevertAssert();
            }

            return changed;
        }


        public bool HasJoinedEntities()
        {
            foreach (PropertyInfo prop in (GetType().GetProperties()))
            {
                if (prop.PropertyType == typeof(EntityCollection))
                {
                    EntityCollection list = (EntityCollection)prop.GetValue(this, null);
                    if (list != null)
                        return true;
                }
            }
            return false;
        }
        /// <summary>
        /// Qualifies a field with the table name or alias.
        /// </summary>
        /// <param name="field">Field to qualify.</param>
        /// <returns>Qualified field.</returns>
        public virtual string QualifyField(string field)
        {
            return GetType().Name + "." + field;
        }

        /// <summary>
        /// Qualifies a field with the table name or alias.
        /// </summary>
        /// <param name="field">Field to qualify.</param>
        /// <returns>Qualified field.</returns>
        public virtual string MapField(string field)
        {
            return field;
        }

        /// <summary>
        /// Gets join for current entity and table.
        /// </summary>
        /// <param name="tableName">Table to join with entity.</param>
        /// <returns>Join definition for entity and table.</returns>
        public virtual DBJoin GetJoin(string tableName)
        {
            return null;
        }

        /// <summary>
        /// Returns the table name for entity.
        /// </summary>
        /// <returns></returns>
        public virtual string TableName()
        {
            if (string.IsNullOrEmpty(_UnderlyingTableName))
                return TableAlias();
            return _UnderlyingTableName;
        }

        /// <summary>
        /// Returns the Alias for the entity. 
        /// </summary>
        /// <returns></returns>
        public string TableAlias()
        {
            return GetType().Name;
        }


        /// <summary>
        /// Checks if the property is in the table or not.
        /// </summary>
        /// <param name="member">Property</param>
        /// <returns>True if the property is not in the table.</returns>
        public static bool IsNotInTable(MemberInfo member)
        {
            foreach (object attribute in member.GetCustomAttributes(typeof(NotInTableAttribute), true))
            {
                if (attribute is NotInTableAttribute)
                {
                    return true;
                }
            }
            return false;
        }

        public string ToXML()
        {
            StringBuilder XMLEntity = new StringBuilder();
            PropertyInfo[] pi = GetType().GetProperties();
            XMLEntity.Append(String.Format(CultureInfo.InvariantCulture, "<{0}>", GetType().Name));
            foreach (PropertyInfo p in pi)
            {
                if (p.PropertyType == typeof(EntityCollection))
                {
                    EntityCollection inner = (EntityCollection)p.GetValue(this, null);
                    if (inner != null)
                    {
                        foreach (Entity en in inner)
                        {
                            if (en != null)
                            {
                                XMLEntity.Append(en.ToXML());
                            }
                        }
                    }
                }
                else
                    if (p.PropertyType.IsSubclassOf(typeof(Entity)))
                    {
                        if (p.GetValue(this, null) != null)
                        {
                            XMLEntity.Append(((Entity)p.GetValue(this, null)).ToXML());
                        }
                    }
                    else
                        //if (! IsNotInTable(p))
                    {
                        if (p.GetValue(this, null) != null)
                            XMLEntity.Append(String.Format(CultureInfo.InvariantCulture, "<{0}>{1}</{0}>", p.Name, p.GetValue(this, null).ToString()));
                    }
            }
            XMLEntity.Append(String.Format(CultureInfo.InvariantCulture, "</{0}>", GetType().Name));
            return XMLEntity.ToString();
        }


        /// <summary>
        /// Uses the properties to validate the Entity's properties.
        /// </summary>
        /// <returns>True if valid; False if not valid.</returns>
        public ArrayList IsValid()
        {
            ArrayList errors = new ArrayList();
            PropertyInfo[] pi = GetType().GetProperties();
            foreach (PropertyInfo p in pi)
            {
                object val = p.GetValue(this, null);
                if (p.GetCustomAttributes(typeof(IsNotNullableAttribute), true).Length > 0)
                {
                    if (null == val)
                        errors.Add(p.Name + " cannot be null.");
                }
                if (null != val)
                {
                    int precision = PropertyHelper.GetPropertyPrecision(p);
                    if (precision <= 0)
                    {
                        precision = PropertyHelper.GetPropertyMaxLength(p);
                    }
                    {
                        if (precision > 0 && precision < val.ToString().Length)
                            errors.Add(p.Name + " has " + val.ToString().Length.ToString(CultureInfo.CurrentCulture) + " characters and the maximum allowed is " + precision.ToString(CultureInfo.CurrentCulture) + ".");
                    }
                }
            }
            return errors;
        }

        /// <summary>
        /// Makes a deep copy of the current entity.
        /// </summary>
        /// <returns>A copy of the current entity.</returns>
        public object CloneJoin()
        {
            Type t = GetType();
            Entity obj = (Entity)Activator.CreateInstance(t);
            obj.SetFlat(IsFlat());
            PropertyInfo[] pi = t.GetProperties();
            foreach (PropertyInfo p in pi)
            {
                if (p.PropertyType == typeof(EntityCollection))
                {
                    EntityCollection inner = (EntityCollection)p.GetValue(this, null);
                    if (inner != null)
                    {
                        EntityCollection join = ((EntityCollection)p.GetValue(this, null));
                        p.SetValue(obj, new EntityCollection(((Entity)join.ElementInstance).CloneEntity()), null);
                    }
                }
                else
                {
                    if (p.PropertyType.IsSubclassOf(typeof(Entity)))
                    {
                        Entity innerEntity = (Entity)p.GetValue(this, null);
                        if (innerEntity != null)
                            p.SetValue(obj, innerEntity.CloneEntity(), null);
                    }
                }

            }
            return obj;
        }

        public object Clone()
        {
            Entity clone = (Entity) CloneEntity();
            clone.originalEntity = originalEntity;
            clone.entityStatus = entityStatus;
            return clone;
        }

        /// <summary>
        /// Makes a deep copy of the current entity.
        /// </summary>
        /// <returns>A copy of the current entity.</returns>
        public object CloneEntity()
        {
            Type t = GetType();
            Entity obj = (Entity)Activator.CreateInstance(t);
            obj.SetFlat(IsFlat());

            PropertyInfo[] pi = t.GetProperties();
            foreach (PropertyInfo p in pi)
            {
                if (p.PropertyType == typeof(EntityCollection))
                {
                    EntityCollection inner = (EntityCollection)p.GetValue(this, null);
                    if (inner != null)
                    {
                        EntityCollection join = ((EntityCollection)p.GetValue(this, null));
                        p.SetValue(obj, new EntityCollection(((Entity)join.ElementInstance).CloneEntity()), null);
                    }
                }
                else
                    try
                    {
                        if (!IsNotInTable(p))
                        {
                            try
                            {
                                FieldInfo fi = t.GetField("_" + p.Name, BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.IgnoreCase);
                                if (fi != null)
                                {
                                    fi.SetValue(obj, fi.GetValue(this));
                                }
                                else
                                {
                                    p.SetValue(obj, p.GetValue(this, null), null);
                                }
                            }
                            catch (System.ArgumentException)
                            {
                            }
                        }
                    }
                    finally
                    {
                    }
            }
            return obj;
        }
        /// <summary>
        /// Parses a property value to the appropriate type.
        /// </summary>
        /// <param name="theValue">Original value</param>
        /// <param name="valueType">Type to parse into</param>
        /// <returns>Parsed value</returns>
        public static object Parse(object theValue, Type valueType)
        {
            if (theValue == null)
                return null;
            string valueAsString = theValue.ToString();
            object assignedValue = theValue;
            if (valueType == typeof(Boolean))
            {
                if (theValue is String)
                {
                    string stringValue = theValue as string;
                    assignedValue = string.Compare(stringValue, "1", false, CultureInfo.CurrentCulture) == 0 || string.Compare(stringValue, "True", false, CultureInfo.CurrentCulture) == 0;
                }
                else if (theValue is Int32 || theValue is Int16 || theValue is Int64)
                    assignedValue = theValue.Equals(1);
            }
            else
                if (valueType == typeof(Byte))
                {
                    assignedValue = Byte.Parse(valueAsString, CultureInfo.InvariantCulture);
                }
                else
                    if (valueType == typeof(Char))
                    {
                        assignedValue = Char.Parse(valueAsString);
                    }
                    else
                        if (valueType == typeof(DateTime))
                        {
                            assignedValue = DateTime.Parse(valueAsString, CultureInfo.InvariantCulture);
                        }
                        else
                            if (valueType == typeof(Decimal))
                            {
                                assignedValue = Decimal.Parse(valueAsString, CultureInfo.InvariantCulture);
                            }
                            else
                                if (valueType == typeof(Double))
                                {
                                    assignedValue = Double.Parse(valueAsString, CultureInfo.InvariantCulture);
                                }
                                else
                                    if (valueType == typeof(Int16))
                                    {
                                        assignedValue = Int16.Parse(valueAsString, CultureInfo.InvariantCulture);
                                    }
                                    else
                                        if (valueType == typeof(Int32))
                                        {
                                            assignedValue = Int32.Parse(valueAsString, CultureInfo.InvariantCulture);
                                        }
                                        else
                                            if (valueType == typeof(Int64))
                                            {
                                                assignedValue = Int64.Parse(valueAsString, CultureInfo.InvariantCulture);
                                            }
                                            else
                                                if (valueType == typeof(SByte))
                                                {
                                                    assignedValue = SByte.Parse(valueAsString, CultureInfo.InvariantCulture);
                                                }
                                                else
                                                    if (valueType == typeof(Single))
                                                    {
                                                        assignedValue = Single.Parse(valueAsString, CultureInfo.InvariantCulture);
                                                    }
                                                    else
                                                        if (valueType == typeof(String))
                                                        {
                                                            assignedValue = System.Convert.ToString(theValue, CultureInfo.InvariantCulture);
                                                        }
                                                        else
                                                            if (valueType == typeof(UInt16))
                                                            {
                                                                assignedValue = UInt16.Parse(valueAsString, CultureInfo.InvariantCulture);
                                                            }
                                                            else
                                                                if (valueType == typeof(UInt32))
                                                                {
                                                                    assignedValue = UInt32.Parse(valueAsString, CultureInfo.InvariantCulture);
                                                                }
                                                                else
                                                                    if (valueType == typeof(UInt64))
                                                                    {
                                                                        assignedValue = UInt64.Parse(valueAsString, CultureInfo.InvariantCulture);
                                                                    }
            return assignedValue;
        }
        public void Dispose()
        {
        }
    }
}